from typing import Union
import numpy as np
import torch

from hbrl.agents.agent import Agent
from abc import ABC, abstractmethod
from gym.spaces import Box, Discrete
from hbrl.agents.utils.replay_buffer import ReplayBuffer


class ValueBasedAgent(Agent, ABC):

    def __init__(self, state_space: Union[Box, Discrete], action_space: Union[Box, Discrete], **params):
        """
        @param state_space: Environment's state space.
        @param action_space: Environment's action_space.
        @param params: Optional parameters.
        """
        super().__init__(state_space, action_space, **params)

        self.batch_size = params.get("batch_size", 100)
        self.buffer_max_size = params.get("buffer_max_size", int(1e6))
        self.replay_buffer = ReplayBuffer(self.buffer_max_size, self.device)

    @abstractmethod
    def get_value(self, features, actions=None):
        return 0

    @abstractmethod
    def learn(self):
        pass

    def scale_action(self, actions: Union[np.ndarray, torch.Tensor], source_action_box: Box):
        """
        Scale an action within the given bounds action_low to action_high, to our action_space.
        The result action is also clipped to fit in the action space in case the given action wasn't exactly inside
        the given bounds.
        Useless if our action space is discrete.
        @return: scaled and clipped actions. WARNING: actions are both attribute and result. They are modified by the
        function. They are also returned for better convenience.
        """
        assert isinstance(self.action_space, Box), \
            "Scale_action is useless and cannot work if our action space is discrete."
        assert isinstance(actions, np.ndarray) or isinstance(actions, torch.Tensor)
        assert isinstance(source_action_box, Box)

        source_low, source_high = source_action_box.low, source_action_box.high
        target_low, target_high = self.action_space.low, self.action_space.high
        if isinstance(actions, torch.Tensor):
            source_low, source_high = torch.tensor(source_low), torch.tensor(source_high)
            target_low, target_high = torch.tensor(target_low), torch.tensor(target_high)

        # Scale action to the action space
        source_range = source_high - source_low
        target_range = target_high - target_low

        scale = target_range / source_range
        actions = actions * scale
        actions = actions + (target_low - (source_low * scale))
        clip_fun = np.clip if isinstance(actions, np.ndarray) else torch.clamp
        actions = clip_fun(actions, target_low, target_high)
        return actions

    def save_interaction(self, *interaction_data):
        """
        Function that is called to ask our agent to learn about the given interaction. This function is separated from
        self.on_action_stop(**interaction_data) because we can imagine agents that do not learn on every interaction, or
        agents that learn on interaction they didn't make (like HER that add interaction related to fake goals in their
        last trajectory).
        on_action_stop is used to update variables likes self.last_state or self.simulation_time_step_id, and
        learn_interaction is used to know the set of interactions we can learn about.

        Example: Our implementation of HER show a call to 'learn_interaction' without 'on_action_stop'
        (two last lines of 'her' file).
        """
        assert not self.under_test
        self.replay_buffer.append(interaction_data)

    def sample_training_batch(self, batch_size=None):
        batch_size = self.batch_size if batch_size is None else batch_size
        return self.replay_buffer.sample(batch_size)

    def process_interaction(self, action, reward, new_state, done, learn=True):
        if learn and not self.under_test:
            self.save_interaction(self.last_state, action, reward, new_state, done)
            self.learn()
        super().process_interaction(action, reward, new_state, done, learn=learn)



import random
import numpy as np
import torch


def to_tensor(data_list, device):
    if isinstance(data_list[0], np.ndarray):
        data_list = np.array(data_list)
    return torch.Tensor(data_list).to(device=device)


class ReplayBuffer:
    def __init__(self, capacity, device, same_size_data=True):
        assert device is not None
        self.capacity = int(capacity)  # capacity of the buffer
        self.device = device
        self.data = []
        self.index = 0  # index of the next cell to be filled
        self.same_size_data = same_size_data
        if self.same_size_data:
            self.data_size = None

    def append(self, data):
        if self.same_size_data and self.data_size is None:
            self.data_size = len(data)
        elif self.same_size_data:
            assert self.data_size == len(data), "Impossible to store data of size " + str(len(data)) + " inside " \
                                                "buffer with data of size " + str(self.data_size) + "."
        assert isinstance(data[0], np.ndarray)

        if len(self.data) < self.capacity:
            self.data.append(None)
        self.data[self.index] = data
        self.index = (self.index + 1) % self.capacity

    def sample(self, batch_size):
        batch_size = min(batch_size, len(self.data))
        batch = random.sample(self.data, batch_size)
        try:
            return list(map(lambda x: to_tensor(x, self.device), list(zip(*batch))))
        except Exception as e:
            result = []
            for elt_id, elt in enumerate(zip(*batch)):
                try:
                    result.append(to_tensor(elt, self.device))
                except Exception as exc:
                    print()
            print()

    def __len__(self):
        return len(self.data)



import pickle
from copy import deepcopy
import numpy as np
import torch
from torch.nn import ReLU
from hbrl.agents.discrete.dqn import DQN
from hbrl.agents.utils.mlp import MLP
from torch import optim
import torch.nn.functional as F

from hbrl.utils import create_dir


class DistributionalDQN(DQN):
    """
    A distributional RL version of DQN, but implemented in the same way that SORB authors describe a
    distributional version of DDPG in SORB's paper (Eysenbach, 2017).
    """

    name = "Distributional DQN"

    def __init__(self, state_space, action_space, **params):
        """
        @param state_space: Environment's state space.
        @param action_space: Environment's action_space.
        @param params: Optional parameters.
        """

        super().__init__(state_space, action_space, **params)
        self.nb_models = params.get("nb_models", 3)
        self.out_dist_size = params.get("output_distribution_size", 20)
        self.out_dist_abscissa = - np.linspace(0, self.out_dist_size, self.out_dist_size)
        self.out_dist_abscissa_steps = self.out_dist_size / (self.out_dist_size - 1)

        model_layers = params.get("model_layers", [64, ReLU(), 64, ReLU(), 64, ReLU()])
        self.models = [
            MLP(self.state_size, model_layers, self.nb_actions * self.out_dist_size, learning_rate=self.learning_rate,
                optimizer_class=optim.Adam, device=self.device).float() for _ in range(self.nb_models)
        ]
        self.target_models = [deepcopy(self.models[i]) for i in range(self.nb_models)]
        self.errors = {}
        for i in range(self.nb_models):
            self.errors["model_" + str(i)] = []

        # Remove the attributes of the mother class that we will not use
        del self.model
        del self.target_model

    def get_q_values_probabilities(self, observations, use_target=False):
        is_batch = len(observations.shape) > 1
        batch_size = observations.shape[0] if is_batch else 1
        models = self.target_models if use_target else self.models
        if is_batch:
            new_shape = (batch_size, self.nb_actions, self.out_dist_size)
        else:
            new_shape = (self.nb_actions, self.out_dist_size)
        q_values_probs = []
        for model_id in range(self.nb_models):
            value = models[model_id](observations)
            value = value.reshape(new_shape)
            value = F.softmax(value, dim=-1)
            q_values_probs.append(value)
        return torch.stack(q_values_probs).mean(0)

    def get_value(self, observations, actions=None, use_target=False):
        with torch.no_grad():
            q_values_probs = self.get_q_values_probabilities(observations, use_target)
            q_values_per_action = (q_values_probs * torch.tensor(self.out_dist_abscissa)).sum(dim=-1)
            if actions is None:
                values = q_values_per_action.max(-1).values
            else:
                if not isinstance(actions, torch.Tensor):
                    actions = torch.tensor(actions)
                values = q_values_per_action.gather(1, actions.to(torch.long).unsqueeze(1))
            return values.detach().numpy()

    def action(self, state, explore=True):
        if not self.under_test:
            if self.simulation_time_step_id > self.epsilon_decay_delay:
                self.epsilon = max(self.epsilon_min, self.epsilon - self.epsilon_step)

        if explore and not self.under_test and np.random.rand() < self.epsilon:  # Epsilon greedy
            action = np.random.randint(self.nb_actions)
        else:
            # greedy_action(self.model, observation) function in RL5 notebook
            with torch.no_grad():
                q_values_probs = self.get_q_values_probabilities(state)
                q_values_per_action = (q_values_probs * torch.tensor(self.out_dist_abscissa)).sum(dim=-1)
                action = q_values_per_action.max(-1).indices.detach().cpu().item()
        return action

    def learn(self):
        assert not self.under_test
        if len(self.replay_buffer) > self.batch_size:
            states, actions, rewards, new_states, dones = self.sample_training_batch()

            with torch.no_grad():
                target_q_values_probs = self.get_q_values_probabilities(new_states, use_target=True)

                target_q_values_per_action = (target_q_values_probs * torch.tensor(self.out_dist_abscissa)).sum(dim=-1)
                next_actions = target_q_values_per_action.max(-1).indices
                next_actions = next_actions.to(torch.long).repeat((self.out_dist_size, 1)).T.unsqueeze(1)
                target_q_values_probabilities = target_q_values_probs.gather(1, next_actions).squeeze()

            # Compute target

            # Compute distribution target
            # - Target distributions for samples where the goal has been reached
            #   shape: [[1, 0, 0, ...],
            #           [1, 0, 0, ...], ... ]

            reached_target_distribution = F.one_hot(torch.zeros(self.batch_size).to(dtype=torch.long),
                                                    self.out_dist_size)

            # - Target distribution for samples where the goal hasn't been reached
            #    * Build a column of 0
            first_column = torch.zeros(self.batch_size).unsqueeze(1)
            #    * Build others columns by shifting the target q_values
            middle_columns = target_q_values_probabilities[:, :-2]
            #    * Build the last column as the sum of the last two left q_values
            last_column = target_q_values_probabilities[:, -2:].sum(-1).unsqueeze(1)
            failed_target_distribution = torch.concat((first_column, middle_columns, last_column), -1)

            # Compute the target distributions
            target_distribution = torch.where(dones[:, np.newaxis] == 1.,
                                              reached_target_distribution,
                                              failed_target_distribution)

            for model_id in range(self.nb_models):
                model = self.models[model_id]
                model_distribution = model(states).reshape((self.batch_size, self.nb_actions, self.out_dist_size))
                a = actions.to(torch.long).repeat((self.out_dist_size, 1)).T.unsqueeze(1)
                model_distribution = model_distribution.gather(1, a).squeeze()
                model_loss = F.cross_entropy(model_distribution, target_distribution)
                model.learn(model_loss)
                loss_copy = model_loss.detach().cpu().item()
                self.errors["model_" + str(model_id)].append(loss_copy)
                self.target_models[model_id].converge_to(model, tau=self.tau)

    def save(self, directory):
        if directory[-1] != "/":
            directory += "/"
        directory = directory[:-1]
        directory += "_" + str(self.nb_models) + "/"
        create_dir(directory)

        for model_id, model in enumerate(self.models):
            torch.save(model, directory + "model_" + str(model_id) + ".pt")

        with open(directory + "replay_buffer.pkl", "wb") as f:
            pickle.dump(self.replay_buffer, f)

    def load(self, directory):
        if directory[-1] != "/":
            directory += "/"
        directory = directory[:-1]
        directory += "_" + str(self.nb_models) + "/"
        create_dir(directory)

        for model_id, model in enumerate(self.models):
            self.models[model_id] = torch.load(directory + "model_" + str(model_id) + ".pt")

        with open(directory + "replay_buffer.pkl", "rb") as f:
            self.replay_buffer = pickle.load(f)



import numpy as np
from .point_env import PointEnv
from .utils.indexes import Colors
from ..maps.maps_index import MapsIndex


class GoalConditionedPointEnv(PointEnv):
    def __init__(self, map_name:str = MapsIndex.EMPTY.value, action_noise=1.0, reset_anywhere=True,
                 reachability_threshold=0.7, dense_reward=False):
        super().__init__(map_name=map_name, action_noise=action_noise, reset_anywhere=reset_anywhere)
        self.goal = None
        self.reachability_threshold = reachability_threshold
        self.dense_reward = dense_reward

    def reset(self) -> tuple:
        """
        Return the initial state, and the selected goal.
        """
        self.goal = self._sample_empty_state().copy()
        return super().reset(), self.goal

    def reached(self, state=None, goal=None):
        """
        Return True if the goal is considered as reached according to the environment reachability threshold.
        The state used is the current agent state if the state parameter if left empty.
        The goal used id the current episode goal if the goal parameter is left empty.
        """
        state = self.agent_state.copy() if state is None else state
        goal = self.goal.copy() if goal is None else goal
        return np.linalg.norm(state - goal) < self.reachability_threshold

    def step(self, action):
        super().step(action)

        done = self.reached()
        if self.dense_reward:
            reward = - np.linalg.norm(self.agent_state - self.goal)
        else:
            reward = 0 if done else -1
        return self.agent_state.copy(), reward, done

    def render(self, ignore_rewards=True, ignore_goal=False):
        """
        Render the whole-grid human view (get view from super class then add the goal over the image)
        """
        img = super().render(ignore_rewards=ignore_rewards)
        if not ignore_goal:
            self.place_point(img, self.goal.copy(), Colors.GOAL.value)
        return img



import importlib
import math
import random
from queue import PriorityQueue
from typing import Any, Tuple, Union, Dict, Optional
from gym import spaces
import numpy as np
from PIL import Image
from scipy.spatial import distance
from .utils.indexes import *
from hbrl.utils.sys_fun import create_dir
from ..maps.maps_index import MapsIndex
from skimage.draw import line_aa


# un environment custom simple
def euclidean_distance(coordinates_1: tuple, coordinates_2: tuple) -> float:
    x1, y1 = coordinates_1
    x2, y2 = coordinates_2
    return math.sqrt((x2 - x1)**2 + (y2 - y1)**2)


class DiscreteGridWorld:
    start_coordinates: Tuple[Union[int, Any], int]
    metadata = {'render.modes': ['human']}

    def __init__(self, map_name:str = MapsIndex.EMPTY.value):

        self.maze_map = np.array(importlib.import_module("hbrl.environments.maps." + map_name).maze_array)
        self.height, self.width = self.maze_map.shape
        self.agent_coordinates = None

        low = np.array([.5 - self.width / 2, - (self.height / 2 - .5)])
        high = np.array([self.width / 2 - .5, - (.5 - self.height / 2)])
        self.state_space = spaces.Box(low=low, high=high)
        self.action_space = spaces.Discrete(len(Direction))
        self.reset()

    def get_state(self, x, y):
        """
        Return a numpy array (state) that belongs to X and Y coordinates in the grid.
        """
        x_value = x + .5 - self.width / 2
        y_value = - (y + .5 - self.height / 2)
        return np.asarray([x_value, y_value])

    def get_coordinates(self, state):
        return round(state[0].item() -.5 + self.width / 2), round(- state[1].item() - .5 + self.height / 2)

    def is_valid_coordinates(self, x, y):
        return 0 <= x < self.width and 0 <= y < self.height

    def is_valid_state(self, state):
        return self.is_valid_coordinates(*self.get_coordinates(state))

    def get_tile_type(self, x, y):
        return TileType(self.maze_map[y][x].item())

    def is_terminal_tile(self, x, y):
        state_type = self.get_tile_type(x, y)
        return state_type == TileType.TERMINAL

    def is_available(self, x, y):
        # False for 218, 138
        # if we move into a row not in the grid
        if 0 > x or x >= self.width or 0 > y or y >= self.height:
            return False
        if self.get_tile_type(x, y) == TileType.WALL:
            return False
        return True

    def get_new_coordinates(self, action):
        agent_x, agent_y = self.agent_coordinates
        if Direction(action) == Direction.TOP:
            agent_y -= 1
        elif Direction(action) == Direction.BOTTOM:
            agent_y += 1
        elif Direction(action) == Direction.LEFT:
            agent_x -= 1
        elif Direction(action) == Direction.RIGHT:
            agent_x += 1
        else:
            raise AttributeError("Unknown action")
        return agent_x, agent_y

    def step(self, action):
        new_x, new_y = self.get_new_coordinates(action)
        if self.is_available(new_x, new_y):
            done = self.is_terminal_tile(new_x, new_y)
            reward = -1 if not done else 1
            self.agent_coordinates = new_x, new_y
            return self.get_state(self.agent_coordinates[0], self.agent_coordinates[1]), reward, done
        else:
            return self.get_state(self.agent_coordinates[0], self.agent_coordinates[1]), -1, False

    def reset(self):
        self.agent_coordinates = np.flip(random.choice(np.argwhere(self.maze_map == 2)))
        state = self.get_state(*self.agent_coordinates)
        return state

    def get_color(self, x, y, ignore_agent=False, ignore_terminals=False):
        agent_x, agent_y = self.agent_coordinates
        if (agent_x, agent_y) == (x, y) and not ignore_agent:
            return Colors.AGENT.value
        else:
            tile_type = self.get_tile_type(x, y)
            if tile_type == TileType.START:
                return Colors.START.value
            elif tile_type == TileType.WALL:
                return Colors.WALL.value
            elif tile_type == TileType.EMPTY:
                return Colors.EMPTY.value
            elif tile_type == TileType.TERMINAL:
                return Colors.EMPTY.value if ignore_terminals else Colors.TERMINAL.value
            else:
                raise AttributeError("Unknown tile type")

    def set_tile_color(self, image_array: np.ndarray, x, y, color, tile_size=10, border_size=0) -> np.ndarray:
        """
        Set a tile color with the given color in the given image as a numpy array of pixels
        :param image_array: The image where the tile should be set
        :param x: X coordinate of the tile to set
        :param y: Y coordinate of the tile to set
        :param color: new color of the tile : numpy array [Red, Green, Blue]
        :param tile_size: size of the tile in pixels
        :param border_size: size of the tile's border in pixels
        :return: The new image
        """
        tile_img = np.zeros(shape=(tile_size, tile_size, 3), dtype=np.uint8)

        if border_size > 0:
            tile_img[:, :, :] = Colors.TILE_BORDER.value
            tile_img[border_size:-border_size, border_size:-border_size, :] = color
        else:
            tile_img[:, :, :] = color

        y_min = y * tile_size
        y_max = (y + 1) * tile_size
        x_min = x * tile_size
        x_max = (x + 1) * tile_size
        image_array[y_min:y_max, x_min:x_max, :] = tile_img
        return image_array

    def get_environment_background(self, tile_size=10, ignore_agent=True, ignore_rewards=False) -> np.ndarray:
        """
        Return an image (as a numpy array of pixels) of the environment background.
        :return: environment background -> np.ndarray
        """
        # Compute the total grid size
        width_px = self.width * tile_size
        height_px = self.height * tile_size

        img = np.zeros(shape=(height_px, width_px, 3), dtype=np.uint8)

        # Render the grid
        for y in range(self.height):
            for x in range(self.width):
                cell_color = self.get_color(x, y, ignore_agent=ignore_agent, ignore_terminals=ignore_rewards)
                img = self.set_tile_color(img, x, y, cell_color)
        return img

    def get_oracle(self) -> list:
        """
        Return an oracle as a list of every possible states inside the environment.
        """
        reachable_coordinates = np.argwhere(self.maze_map != 1).tolist()
        return [self.get_state(x, y) for x, y in reachable_coordinates]

    def render(self, ignore_rewards=False):
        """
        Render the whole-grid human view
        """
        if self.agent_coordinates is None:
            self.reset()
        img = self.get_environment_background(ignore_agent=False, ignore_rewards=ignore_rewards)
        agent_x, agent_y = self.agent_coordinates
        self.place_point(img, self.get_state(agent_x, agent_y), Colors.AGENT.value)
        return img

    def get_available_positions(self, coordinates: tuple) -> list:
        """
        return an list of every available coordinates from the given one (used for A*).
        """
        x, y = coordinates  # Make sure coordinates is a tuple

        available_coordinates = []
        if x < (self.width - 1):
            new_coord = (x + 1, y)
            if self.is_available(x + 1, y):
                available_coordinates.append((new_coord, Direction.RIGHT.value))
        if x > 0:
            new_coord = (x - 1, y)
            if self.is_available(x - 1, y):
                available_coordinates.append((new_coord, Direction.LEFT.value))

        if y < (self.height - 1):
            new_coord = (x, y + 1)
            if self.is_available(x, y + 1):
                available_coordinates.append((new_coord, Direction.BOTTOM.value))
        if y > 0:
            new_coord = (x, y - 1)
            if self.is_available(x, y - 1):
                available_coordinates.append((new_coord, Direction.TOP.value))

        return available_coordinates

    def best_path(self, state_1, state_2):
        """
        Return the shortest distance between two tiles, in number of action the agent needs to go from one to another.
        :param state_1: Start state,
        :param state_2: Destination state,
        :return: Shortest path (using A*), as a list of action to move from state_1 to state_2.
        """
        # Remove trivial case
        if isinstance(state_1, tuple):
            coordinates_1 = state_1
        else:
            coordinates_1 = self.get_coordinates(state_1)
        if isinstance(state_2, tuple):
            coordinates_2 = state_2
        else:
            coordinates_2 = self.get_coordinates(state_2)

        frontier = PriorityQueue()
        frontier.put((0, coordinates_1))
        came_from: Dict[tuple, Optional[tuple]] = {}
        cost_so_far = {coordinates_1: 0}
        came_from[coordinates_1] = None

        while not frontier.empty():
            priority, current = frontier.get()

            if current == coordinates_2:
                break

            for next_position, action in self.get_available_positions(current):
                new_cost = cost_so_far[current] + 1
                if next_position not in cost_so_far or new_cost < cost_so_far[next_position]:
                    cost_so_far[next_position] = new_cost
                    priority = new_cost + euclidean_distance(next_position, coordinates_2)
                    frontier.put((priority, next_position))
                    came_from[next_position] = current

        return came_from, cost_so_far

    def distance(self, state_1, state_2):
        _, distance = self.best_path(state_1, state_2)
        coordinates_2 = self.get_coordinates(state_2)
        return distance[coordinates_2]

    def show_path_on_image(self, state_1, state_2, file_directory, file_name, colors) -> None:
        """
        Save an image of the environment with many path draw on it
        :param state_1: start state,
        :param state_2: destination state,
        :param file_directory: destination where the image should be saved
        :param file_name: name of the future image (without .png)
        :param colors: list of colors of shape [[R1, G1, B1], [R1, G2, B2], ... ] and len = len(paths)
        """

        best_path = self.best_path(state_1, state_2)

        if isinstance(state_1, tuple):
            coordinates_1 = state_1
        else:
            coordinates_1 = self.get_coordinates(state_1)
        if isinstance(state_2, tuple):
            coordinates_2 = state_2
        else:
            coordinates_2 = self.get_coordinates(state_2)

        if not self.is_available(coordinates_1[0], coordinates_1[1]) or not self.is_available(coordinates_2[0],
                                                                                              coordinates_2[1]):
            print("one of these states is not available")

        # Generate image
        image = self.get_environment_background()

        # Draw this path
        for coordinates in best_path.coordinates:
            tile_x, tile_y = coordinates
            self.set_tile_color(image, tile_x, tile_y, colors)

        # Save image
        image = Image.fromarray(image)
        create_dir(file_directory)
        if not file_name.endswith(".png"):
            if len(file_name.split(".")) > 1:
                file_name = "".join(file_name.split(".")[:-1])  # Remove the last extension
            file_name += ".png"
        image.save(file_directory + file_name)

    def sample_reachable_state(self):
        state_coordinates = np.flip(random.choice(np.argwhere(self.maze_map != 1)))
        return self.get_state(*state_coordinates)

    def place_point(self, image: np.ndarray, state, color: Union[np.ndarray, list], width=5):
        """
        Modify the input image
        param image: Initial image that will be modified.
        param x: x coordinate in the state space of the point to place.
        param y: y coordinate in the state space of the point to place.
        param color: Color to give to the pixels that compose the point.
        param width: Width of the circle (in pixels).
        """
        if isinstance(color, list):
            color = np.array(color)

        center_x, center_y = self.get_coordinates(state)
        center_x = (center_x + 0.5) / self.width
        center_y = (center_y + 0.5) / self.height
        center_y, center_x = (image.shape[:2] * np.array([center_y, center_x])).astype(int)

        # Imagine a square of size width * width, with the coordinates computed above as a center. Iterate through
        # each pixel inside this square to
        radius = round(width / 2) + 1
        for i in range(center_x - radius, center_x + radius):
            for j in range(center_y - radius, center_y + radius):
                if distance.euclidean((i + 0.5, j + 0.5), (center_x, center_y)) < radius:
                    image[j, i] = color

        return image

    def place_edge(self, image: np.ndarray, state_1, state_2, color: Union[np.ndarray, list], width=40):
        """
        Modify the input image
        param image: Initial image that will be modified.
        param x: x coordinate in the state space of the point to place.
        param y: y coordinate in the state space of the point to place.
        param color: Color to give to the pixels that compose the point.
        param width: Width of the circle (in pixels).
        """

        color = np.array(color) if isinstance(color, list) else color
        center_x_1, center_y_1 = self.get_coordinates(state_1)
        center_x_1 = (center_x_1 + 0.5) / self.width
        center_y_1 = (center_y_1 + 0.5) / self.height
        center_y_1, center_x_1 = (image.shape[:2] * np.array([center_y_1, center_x_1])).astype(int)

        center_x_2, center_y_2 = self.get_coordinates(state_2)
        center_x_2 = (center_x_2 + 0.5) / self.width
        center_y_2 = (center_y_2 + 0.5) / self.height
        center_y_2, center_x_2 = (image.shape[:2] * np.array([center_y_2, center_x_2])).astype(int)

        rr, cc, val = line_aa(center_y_1, center_x_1, center_y_2, center_x_2)
        old = image[rr, cc]
        extended_val = np.tile(val, (3, 1)).T
        image[rr, cc] = (1 - extended_val) * old + extended_val * color

        return image



import random

import numpy as np

from .discrete_grid_world import DiscreteGridWorld
from ..maps.maps_index import MapsIndex
from .utils.indexes import Colors


class GoalConditionedDiscreteGridWorld(DiscreteGridWorld):
    def __init__(self, map_name: str = MapsIndex.EMPTY.value):
        super().__init__(map_name=map_name)
        self.goal_coordinates = None
        self.goal = None
        self.reachability_threshold = 0.1  # In this environment, an (s - g) L2 norm below this threshold implies s = g

    def reset_goal(self):
        """
        Choose a goal for the agent.
        """
        self.goal_coordinates = np.flip(random.choice(np.argwhere(self.maze_map != 1)))
        self.goal = self.get_state(*self.goal_coordinates)

    def goal_reached(self):
        """
        Return a boolean True if the agent state is on the goal (and exactly on the goal since our state space is
        discrete here in reality), and false otherwise.
        """
        return (self.agent_coordinates == self.goal_coordinates).all()

    def step(self, action):
        new_x, new_y = self.get_new_coordinates(action)
        if self.is_available(new_x, new_y):
            self.agent_coordinates = new_x, new_y
            done = self.goal_reached()
            reward = -1 if not done else 0
            return self.get_state(self.agent_coordinates[0], self.agent_coordinates[1]), reward, done
        else:
            return self.get_state(self.agent_coordinates[0], self.agent_coordinates[1]), -1, False

    def reset(self) -> tuple:
        """
        Return the initial state, and the selected goal.
        """
        self.reset_goal()
        return super().reset(), self.goal

    def render(self, ignore_rewards=True):
        """
        Render the whole-grid human view (get view from super class then add the goal over the image)
        """
        if self.goal_coordinates is None:
            self.reset()
        img = super().render(ignore_rewards=ignore_rewards)
        goal_x, goal_y = self.goal_coordinates
        self.place_point(img, self.get_state(goal_x, goal_y), Colors.GOAL.value)
        return img



import abc
import importlib
import random
from typing import Union
from scipy.spatial import distance
from skimage.draw import line_aa
import numpy as np
from gym.spaces import Box, Discrete
from .environment import Environment
from .maps.maps_index import MapsIndex
from enum import Enum

from hbrl.environments.goal_reaching_environment import GoalReachingEnv


class TileType(Enum):
    EMPTY = 0
    WALL = 1
    START = 2
    TERMINAL = 3


class Colors(Enum):
    EMPTY = [250, 250, 250]
    WALL = [50, 54, 51]
    START = [213, 219, 214]
    TERMINAL = [73, 179, 101]
    TILE_BORDER = [50, 54, 51]
    AGENT = [0, 0, 255]
    GOAL = [255, 0, 0]


class NavigationEnv(Environment, abc.ABC):

    """
    An environment where the agent navigate inside a maze.
    It will load a map and is able to render image that show a maze representation.
    """

    name = "Default navigation environment"

    def __init__(self, map_tag: MapsIndex, state_space: Union[Box, Discrete], action_space: Union[Box, Discrete],
                 reset_anywhere=False):

        self.maze_map = np.array(importlib.import_module("hbrl.environments.maps." + map_tag.value).maze_array)
        self.height, self.width = self.maze_map.shape
        self.state = None
        self.reset_anywhere = reset_anywhere

        high = np.array([self.width, self.height]) / 2
        low = - high

        self.state_space = Box(low=np.concatenate((low, state_space.low)),
                               high=np.concatenate((high, state_space.high)))
        self.action_space = action_space

        super().__init__(self.state_space, self.action_space)

    def get_state_from_coordinates(self, x, y, uniformly_sampled=True):
        """
        Return a numpy array (state) that belongs to X and Y coordinates in the grid.
        If uniformly_sampled == True, the result state is a state uniformly sampled in the area
        covered by tile at coordinate x, y.
        """
        assert self.is_valid_coordinates(x, y)
        x_value = x + .5 - self.width / 2
        y_value = - (y + .5 - self.height / 2)
        if uniformly_sampled:
            noise = np.random.uniform(-0.5, 0.5, (2,))
        else:
            noise = np.zeros(2)
        return np.asarray([x_value, y_value]) + noise

    def get_coordinates(self, state):
        """
        return the coordinates of the tile where the given state is inside.
        """
        return round(state[0].item() -.5 + self.width / 2), round(- state[1].item() - .5 + self.height / 2)

    def is_valid_coordinates(self, x, y):
        return 0 <= x < self.width and 0 <= y < self.height

    def is_valid_state(self, state):
        return self.state_space.contains(state.astype(self.state_space.dtype))

    def reachable(self, state):
        return self.is_valid_state(state) and self.get_tile_type(*self.get_coordinates(state)) != TileType.WALL

    def get_tile_type(self, x: int, y: int) -> TileType:
        assert self.is_valid_coordinates(x, y)
        return TileType(self.maze_map[y][x].item())

    def is_terminal_tile(self, x: int, y: int) -> bool:
        assert self.is_valid_coordinates(x, y)
        return self.get_tile_type(x, y) == TileType.TERMINAL

    def get_reward_done(self, state) -> (float, bool):
        done = self.is_terminal_tile(*self.get_coordinates(state))
        return -1 if not done else 1, done

    @abc.abstractmethod
    def step(self, action):
        pass

    def reset(self):
        if self.reset_anywhere:
            self.state = self.sample_reachable_state()
        else:
            start_tile = np.flip(random.choice(np.argwhere(self.maze_map == 2)))
            self.state = self.get_state_from_coordinates(*start_tile, uniformly_sampled=False)
        return self.state

    """
    Rendering functions
    """
    def get_color(self, x, y, ignore_agent=False, ignore_terminals=False):
        agent_x, agent_y = self.get_coordinates(self.state)
        if (agent_x, agent_y) == (x, y) and not ignore_agent:
            return Colors.AGENT.value
        else:
            tile_type = self.get_tile_type(x, y)
            if tile_type == TileType.START:
                return Colors.START.value
            elif tile_type == TileType.WALL:
                return Colors.WALL.value
            elif tile_type == TileType.EMPTY:
                return Colors.EMPTY.value
            elif tile_type == TileType.TERMINAL:
                return Colors.EMPTY.value if ignore_terminals else Colors.TERMINAL.value
            else:
                raise AttributeError("Unknown tile type")

    def set_tile_color(self, image_array: np.ndarray, x, y, color, tile_size=10, border_size=0) -> np.ndarray:
        """
        Set a tile color with the given color in the given image as a numpy array of pixels
        :param image_array: The image where the tile should be set
        :param x: X coordinate of the tile to set
        :param y: Y coordinate of the tile to set
        :param color: new color of the tile : numpy array [Red, Green, Blue]
        :param tile_size: size of the tile in pixels
        :param border_size: size of the tile's border in pixels
        :return: The new image
        """
        tile_img = np.zeros(shape=(tile_size, tile_size, 3), dtype=np.uint8)

        if border_size > 0:
            tile_img[:, :, :] = Colors.TILE_BORDER.value
            tile_img[border_size:-border_size, border_size:-border_size, :] = color
        else:
            tile_img[:, :, :] = color

        y_min = y * tile_size
        y_max = (y + 1) * tile_size
        x_min = x * tile_size
        x_max = (x + 1) * tile_size
        image_array[y_min:y_max, x_min:x_max, :] = tile_img
        return image_array

    def get_environment_background(self, tile_size=10, ignore_rewards=False) -> np.ndarray:
        """
        Return an image (as a numpy array of pixels) of the environment background.
        :return: environment background -> np.ndarray
        """
        # Compute the total grid size
        width_px = self.width * tile_size
        height_px = self.height * tile_size

        img = np.zeros(shape=(height_px, width_px, 3), dtype=np.uint8)

        # Render the grid
        for y in range(self.height):
            for x in range(self.width):
                cell_color = self.get_color(x, y, ignore_agent=True, ignore_terminals=ignore_rewards)
                img = self.set_tile_color(img, x, y, cell_color)
        return img

    def get_oracle(self) -> list:
        """
        Return an oracle as a list of every reachable states inside the environment.
        """
        reachable_coordinates = np.argwhere(self.maze_map != 1).tolist()
        return [self.get_state_from_coordinates(x, y) for x, y in reachable_coordinates]

    def sample_reachable_state(self):
        start_tile = np.flip(random.choice(np.argwhere(self.maze_map != 1)))
        return self.get_state_from_coordinates(*start_tile)

    def render(self, ignore_agent=False, ignore_rewards=False):
        """
        Render the whole-grid human view
        """
        if self.state is None:
            self.reset()
        img = self.get_environment_background(ignore_rewards=ignore_rewards)
        if not ignore_agent:
            self.place_point(img, self.state, Colors.AGENT.value)
        return img

    def place_point(self, image: np.ndarray, state, color: Union[np.ndarray, list], width=5):
        """
        Modify the input image
        param image: Initial image that will be modified.
        param x: x coordinate in the state space of the point to place.
        param y: y coordinate in the state space of the point to place.
        param color: Color to give to the pixels that compose the point.
        param width: Width of the circle (in pixels).
        """
        assert isinstance(state, np.ndarray)
        assert len(state.shape) == 1
        state = state[:2]

        if isinstance(color, list):
            color = np.array(color)

        state_space_range = (self.state_space.high - self.state_space.low)[:2]
        center = (state - self.state_space.low[:2]) / state_space_range
        center[1] = 1 - center[1]
        center_y, center_x = (image.shape[:2] * np.flip(center)).astype(int)

        # Imagine a square of size width * width, with the coordinates computed above as a center. Iterate through
        # each pixel inside this square to
        radius = width
        for i in range(center_x - radius, center_x + radius):
            for j in range(center_y - radius, center_y + radius):
                dist = distance.euclidean((i, j), (center_x, center_y))
                if dist < radius and 0 <= i < image.shape[1] and 0 <= j < image.shape[0]:
                    image[j, i] = color
        return image

    def place_edge(self, image: np.ndarray, state_1, state_2, color: Union[np.ndarray, list], width=40):
        """
        Modify the input image
        param image: Initial image that will be modified.
        param x: x coordinate in the state space of the point to place.
        param y: y coordinate in the state space of the point to place.
        param color: Color to give to the pixels that compose the point.
        param width: Width of the circle (in pixels).
        """
        assert isinstance(state_1, np.ndarray)
        assert len(state_1.shape) == 1
        assert isinstance(state_2, np.ndarray)
        assert len(state_2.shape) == 1
        state_1 = state_1[:2]
        state_2 = state_2[:2]
        if isinstance(color, list):
            color = np.array(color)
        state_space_range = (self.state_space.high - self.state_space.low)[:2]

        center = (state_1 - self.state_space.low[:2]) / state_space_range
        center[1] = 1 - center[1]
        center_y_1, center_x_1 = (image.shape[:2] * np.flip(center)).astype(int)

        center = (state_2 - self.state_space.low[:2]) / state_space_range
        center[1] = 1 - center[1]
        center_y_2, center_x_2 = (image.shape[:2] * np.flip(center)).astype(int)

        rr, cc, val = line_aa(center_y_1, center_x_1, center_y_2, center_x_2)
        old = image[rr, cc]
        extended_val = np.tile(val, (3, 1)).T
        image[rr, cc] = (1 - extended_val) * old + extended_val * color


class GoalReachingNavEnv(GoalReachingEnv, NavigationEnv):

    """
    This environment wrap an hbrl.environment.NavigationEnvironment instance to make it goal-conditioned.
    at each reset, a goal is sampled from reachable states in the NavigationEnvironment map, and the reward depends if
    the sampled goal has been reached or not.
    """

    name = "GoalReachingNavEnv Default"

    def __init__(self, wrapped_environment, goal_space: Union[None, Box, Discrete]=None, sparse_reward=True):
        assert isinstance(wrapped_environment, NavigationEnv)
        GoalReachingEnv.__init__(self, wrapped_environment, goal_space, sparse_reward)

    def sample_goal(self):
        goal_coordinates = np.flip(random.choice(np.argwhere(self.maze_map != 1)))
        return self.get_state_from_coordinates(*goal_coordinates, uniformly_sampled=True)

    """
    Rendering functions
    """

    def render(self, ignore_agent=False, ignore_goal=False):
        """
        Render the whole-grid human view
        """
        image = self.wrapped_environment.render(ignore_agent=ignore_agent, ignore_rewards=True)
        if not ignore_goal:
            self.place_point(image, self.goal, Colors.GOAL.value)
        return image


from moviepy.video.io.ImageSequenceClip import ImageSequenceClip
from hbrl.utils.sys_fun import create_dir


def get_red_green_color(value):
    """
    Retourne une couleur correspondant à un gradient entre rouge (0) et vert (1) pour une valeur donnée entre 0 et 1
    :param value: valeur entre 0 et 1 définissant la couleur à récupérer
    """
    value = value * 2 - 1
    if value >= 0:
        red = hex(int((1 - value) * 255))[2:]
        green = hex(255)[2:]
        blue = hex(int((1 - value) * 255))[2:]
    else:
        red = hex(255)[2:]
        green = hex(int((1 + value) * 255))[2:]
        blue = hex(int((1 + value) * 255))[2:]
    red = "0" + red if len(red) == 1 else red
    green = "0" + green if len(green) == 1 else green
    blue = "0" + blue if len(blue) == 1 else blue
    res = "#" + red + green + blue
    return res


def generateVideo(images, directory, filename):
    if len(filename) < 4 or filename[-4:] != ".mp4":
        filename += ".mp4"

    create_dir(directory)
    video = ImageSequenceClip(images, fps=30)
    video.write_videofile(directory + filename)


from enum import Enum


class MapsIndex(Enum):
    EMPTY = "empty_room"
    FOUR_ROOMS = "four_rooms"
    HARD = "hard_maze"
    MEDIUM = "medium_maze"
    EXTREME = "extreme_maze"
    IMPOSSIBLE = "impossible_maze"
    SORB_4_ROOMS = "four_rooms_sorb"
    JOIN_ROOMS = "join_rooms_medium"





class A:
    def __init__(self):
        self.a = "a"

class B(A):
    def __init__(self):
        super().__init__()
        self.b = "b"

class C:
    def __init__(self):
        self.c = "c"

class D(B, C):
    def __init__(self):
        self.d = "d"
        B.__init__(self)

if __name__ == "__main__":
    d = D()
    for attr_name, attr_value in d.__dict__.items():
        print(attr_name, ": ", attr_value)



"""
GWR is a topological graph building strategy, that build a graph that fit the reachability graph of a dataset.
https://www.sciencedirect.com/science/article/pii/S0893608002000783
"""
import random
from typing import Union

import numpy as np
import torch
from gym.spaces import Box, Discrete
from hbrl.agents.global_planning.reachability_graph_learning.rgl import RGL
from hbrl.agents.utils.mlp import MLP
from torch import optim
from torch.nn import ReLU, Sigmoid


class TC_RGL(RGL):

    name = "TC-RGL"

    def __init__(self, goal_conditioned_wrapper, value_based_agent_class,
                 state_space: Union[Box, Discrete], action_space: Union[Box, Discrete], **params):
        """
        @param goal_conditioned_wrapper: Wrapper used to turn the given value based agent class into a goal
            conditioned agent.
        @param value_based_agent_class: Value based agent clas
        @param state_space: Environment's state space.
        @param action_space: Environment's action_space.
        @param params: Optional parameters.
        """

        params["re_usable_policy"] = True
        super().__init__(goal_conditioned_wrapper, value_based_agent_class, state_space, action_space, **params)
        self.ti_tc_network = params.get("translation_invariant_tc_network", True)
        assert isinstance(self.ti_tc_network, bool)
        self.tc_layer_1_size = params.get("tc_layer_1_size", 125)
        self.tc_layer_2_size = params.get("tc_layer_2_size", 100)
        self.tc_learning_rate = params.get("tc_learning_rate", 0.001)
        self.tc_batch_size = params.get("tc_batch_size", 250)
        self.tc_buffer_max_size = params.get("tc_buffer_max_size", 1e9)
        self.nb_tc_data_seen = 0
        self.tc_criterion = params.get("tc_criterion", torch.nn.MSELoss())

        self.targeted_edge_length = params.get("targeted_edge_length", 20)  # Hyperparameter k in stc paper

        self.topological_cognition_network = MLP(self.state_size if self.ti_tc_network else self.state_size * 2,
                                                 self.tc_layer_1_size, ReLU(),
                                                 self.tc_layer_2_size, ReLU(),
                                                 1, Sigmoid(),
                                                 learning_rate=self.tc_learning_rate, optimizer_class=optim.Adam,
                                                 device=self.device).float()

        # We use to replay buffers to have the same amount of positive / negative data in a batch, to limit over-fitting
        self.tc_replay_buffer_positive = []  # Replay buffer for samples with label = 1
        self.tc_replay_buffer_positive = []  #                            ... label = 0
        self.tc_replay_buffer_negative = []
        self.last_episode_trajectory = []

        self.tc_errors_memory = []
        self.tc_average_errors_memory = []
        self.label_0_values = {}
        self.label_1_values = {}

        super().__init__(goal_conditioned_wrapper, value_based_agent_class, state_space, action_space, **params)

    def on_pre_training_done(self, start_state, reached_goals):
        """
        Compute the longer distance estimation over every goal that has been reached during the pre-training.
        It allows to choose reachability parameters more easily.
        """
        pass

    def get_distance_estimation(self, states_batch, goals_batch):
        """
        Return a boolean reflecting the distance of two states according to the TC network.
        """
        if isinstance(states_batch, np.ndarray):
            states_batch = torch.from_numpy(states_batch).to(self.device)
        if isinstance(goals_batch, np.ndarray):
            goals_batch = torch.from_numpy(goals_batch).to(self.device)
        if self.ti_tc_network:
            input_ = torch.concat((states_batch[:, :self.goal_size] - goals_batch[:, :self.goal_size],
                                   states_batch[:, self.goal_size:]), dim=-1)
        else:
            input_ = torch.concat((states_batch, states_batch), dim=-1)
        with torch.no_grad():
            return self.topological_cognition_network(input_).cpu().detach().numpy()

    def get_normalised_distance_estimation(self, states, goals):
        """
        Use the UVFA to get a value function approximation between two states.
        """

        # Convert inputs to batch
        if len(states.shape) == 1:
            states = states[np.newaxis]
        if len(goals.shape) == 1:
            goals = goals[np.newaxis]

        # Operate projections: States should be in the state space and goals in the sub-goal space.
        # - Inverse projection G -> S if the states don't have the right size
        #   (computing distance between goal and state).
        if states.shape[-1] != self.state_size:
            if states.shape[-1] != self.goal_size:
                raise ValueError("Unknown state shape.")
            states = self.goal_to_state(states)

        # - Projection S -> G if the given sub-goals are actually states or goals. Our control policy only know
        #   goals, not goals.
        if goals.shape[-1] != self.goal_size:
            if goals.shape != self.state_size:
                raise ValueError("Unknown goal shape")
            goals = goals[self.state_to_goal_filter]

        # Now inputs have the right size. We can compute the distance.
        # Return the normalised estimated distance (assuming self.distance_estimation_min == 0)
        return self.get_distance_estimation(states, goals).flatten()

    def store_tc_training_samples(self, last_trajectory):
        """
        Use self.last_episode_trajectory to generate and store training samples for the TC network
        """

        for sample_id in range(len(last_trajectory) // 2):

            # Compute data index in the buffer using reservoir sampling
            state_1_index = random.randint(0, len(last_trajectory) - 1)
            state_2_index = random.randint(0, len(last_trajectory) - 1)
            distance = abs(state_2_index - state_1_index)
            state_1 = last_trajectory[state_1_index]
            state_2 = last_trajectory[state_2_index]
            label = 0 if distance < self.targeted_edge_length else 1
            if label == 0:
                if len(self.tc_replay_buffer_negative) > self.tc_buffer_max_size // 2:
                    self.tc_replay_buffer_negative.pop(0)
                self.tc_replay_buffer_negative.append((state_1, state_2, label))
            elif label == 1:
                if len(self.tc_replay_buffer_positive) > self.tc_buffer_max_size // 2:
                    self.tc_replay_buffer_positive.pop(0)
                self.tc_replay_buffer_positive.append((state_1, state_2, label))

        self.train_tc_network()

    def train_tc_network(self):
        if len(self.tc_replay_buffer_positive) > self.tc_batch_size // 2 \
                and len(self.tc_replay_buffer_negative) > self.tc_batch_size // 2:

            # Sample batch data
            p_states_1, p_states_2, p_labels = \
                list(zip(*random.sample(self.tc_replay_buffer_positive, self.tc_batch_size // 2)))
            n_states_1, n_states_2, n_labels = \
                list(zip(*random.sample(self.tc_replay_buffer_negative, self.tc_batch_size // 2)))
            states_1 = p_states_1 + n_states_1
            states_2 = p_states_2 + n_states_2
            labels = p_labels + n_labels
            states_1 = torch.from_numpy(np.array(states_1)).to(self.device)
            states_2 = torch.from_numpy(np.array(states_2)).to(self.device)
            if self.ti_tc_network:
                inputs = torch.concat((states_1[:, :self.goal_size] - states_2[:, :self.goal_size],
                                       states_1[:, self.goal_size:]), dim=-1)
            else:
                inputs = torch.concat((states_1, states_2), dim=-1)

            # Predict label and learn loss
            predictions = self.topological_cognition_network(inputs).squeeze()
            labels = torch.Tensor(list(labels)).to(device=self.device, dtype=torch.float32)
            error = self.tc_criterion(predictions, labels)

            self.topological_cognition_network.learn(error)



from typing import Union

import torch
import numpy as np
import networkx as nx
from copy import deepcopy
from random import choice

from networkx import NetworkXNoPath
from hbrl.agents.agent import Agent
from ...value_based_agent import ValueBasedAgent
from ...goal_conditioned_wrappers.tilo import TILO
from gym.spaces import Discrete, Box


class REO_RGL(Agent):

    """
    REO-RGL stands for Reachability Evaluation Only - Reachability Graph Learning
    This agent is a variant of RGL which get a graph by sampling nodes in its environment. We implement it to observe
    what is the difference between our RGL algorithm, and show that such free information about the environment do not
    present a gain in speed, and reduce the robustness of our algorithm because of a lower nodes quality.
    """

    name = "REO-RGL"

    def __init__(self, goal_conditioned_wrapper, value_based_agent_class,
                 state_space: Union[Box, Discrete], action_space: Union[Box, Discrete], **params):
        """
        @param goal_conditioned_wrapper: Wrapper used to turn the given value based agent class into a goal
            conditioned agent.
        @param value_based_agent_class: Value based agent class.
        @param state_space: Environment's state space.
        @param action_space: Environment's action_space.
        @param params: Optional parameters.
        """

        super().__init__(state_space, action_space, **params)

        assert issubclass(goal_conditioned_wrapper, TILO)
        assert issubclass(value_based_agent_class, ValueBasedAgent)
        self.control_policy_goal_conditioned_wrapper = goal_conditioned_wrapper
        self.control_policy_value_based_agent_class = value_based_agent_class

        # Compute out goal space
        self.goal_space = params.get("goal_space", self.state_space)
        assert isinstance(self.goal_space, Box) or isinstance(self.goal_space, Discrete)
        self.goal_size = self.goal_space.shape[0]
        self.goal_shape = self.goal_space.shape
        assert len(self.goal_shape) == 1, "Multi dimensional spaces are not supported."

        # Compute state_to_goal_filter, (state[self.state_to_goal_filter] = <goal associated to state>) aka a projection
        default_filter = np.array([1] * self.goal_size + [0] * (self.state_size - self.goal_size)).astype(bool)
        self.state_to_goal_filter = params.get("state_to_goal_filter", default_filter)
        assert np.argwhere(self.state_to_goal_filter).shape[0] == self.goal_size

        self.nb_nodes = params.get("nb_nodes", 200)
        self.reachability_graph = nx.Graph()
        self.final_goal = None
        self.control_policy: TILO = goal_conditioned_wrapper(value_based_agent_class, state_space=self.state_space,
                                                             action_space=self.action_space, **params)
        self.done = False
        self.default_state = params.get("default_state")
        assert self.default_state is not None

        self.q_distance_max = None
        self.sub_goals = []

        self.tolerance_radius = params.get("tolerance_radius", 1)

        # Distance estimation will be normalised (like in SORB) between 0 (same state) and 1
        # (The farthest goal observed and reached during pretraining)
        self.edges_distance_threshold = params.get("edges_distance_threshold", 0.2)

        # How many interaction max before we consider that we failed to reach the next sub-goal
        self.max_interactions_per_edge = params.get("max_interactions_per_edge", 40)
        self.current_edge_interactions = 0

        # How many interaction max before we consider that we failed to reach the final goal from the last sub-goal
        self.max_final_interactions = params.get("max_final_interactions", self.max_interactions_per_edge)

        self.sub_goal_size = params.get("sub_goal_size", 2)

        self.verbose = params.get("verbose", False)
        self.last_node_reached = None
        self.reachability_graph_set = False

        self.current_goal = None
        self.state_to_goal_filter = [True for _ in range(self.sub_goal_size)] \
            + [False for _ in range(self.state_size - self.sub_goal_size)]

    def goal_to_state(self, goals):
        goals_ = goals[np.newaxis] if len(goals.shape) == 1 else goals
        states = np.tile(self.default_state, (goals_.shape[0], 1))
        states[:, self.state_to_goal_filter] = goals_
        return goals_

    def init_reachability_graph(self, oracle):
        """
        Initialise a reachability_graph by sampling states in an oracle.
        Param oracle: A list of states that cover the reachable state space.
        """
        self.reachability_graph_set = True
        for i in range(self.nb_nodes):
            self.reachability_graph.add_node(i, state=deepcopy(choice(oracle)))

        nodes_goals = np.array(list(nx.get_node_attributes(self.reachability_graph, "state").values()))
        nodes_states = self.goal_to_state(nodes_goals)
        nodes_states = np.repeat(nodes_states, self.nb_nodes, axis=0)
        nodes_goals = np.tile(nodes_goals, (self.nb_nodes, 1))

        forward_distance_estimations = self.get_distance_estimation(nodes_states.copy(), nodes_goals.copy(),
                                                                    normalised=True)
        backward_distance_estimations = self.get_distance_estimation(nodes_states.copy(), nodes_goals.copy(),
                                                                     normalised=True)

        for first_node, first_attributes in self.reachability_graph.nodes(data=True):
            for second_node, second_attributes in self.reachability_graph.nodes(data=True):

                forward_estimated_distance = forward_distance_estimations[first_node * self.nb_nodes + second_node]
                backward_estimated_distance = backward_distance_estimations[second_node * self.nb_nodes + first_node]
                assert forward_estimated_distance >= 0
                assert backward_estimated_distance >= 0
                if forward_estimated_distance <= self.edges_distance_threshold:
                    self.reachability_graph.add_edge(second_node, first_node, cost=forward_estimated_distance)
                if backward_estimated_distance <= self.edges_distance_threshold:
                    self.reachability_graph.add_edge(first_node, second_node, cost=backward_estimated_distance)

    def init_path(self, state, goal):
        start_node = self.get_node_for_state(state)
        final_node = self.get_node_for_state(goal)
        self.sub_goals = nx.shortest_path(self.reachability_graph, start_node, final_node, "cost")

    def get_node_for_state(self, state, data=False, reachable_only=True):
        """
        Select the node that best represent the given state.
        """

        assert isinstance(state, np.ndarray)
        if state.shape[-1] == len(self.state_to_goal_filter):
            state = state[self.state_to_goal_filter]
        if not self.reachability_graph.nodes:
            return None  # The reachability_graph  hasn't been initialised yet.
        node_data = None
        closest_distance = None
        for node_id, args in self.reachability_graph.nodes(data=True):
            if reachable_only:
                try:  # Try to reach the node
                    nx.shortest_path(self.reachability_graph, 0, node_id)
                except NetworkXNoPath:
                    continue  # Not reachable, inspect the next one.
            distance = self.get_distance_estimation(args["state"], state, normalised=False)
            if closest_distance is None or distance < closest_distance:
                node_data = (node_id, args)
                closest_distance = distance
        return node_data if data else node_data[0]

    def start_episode(self, state, goal, test_episode=False):
        assert isinstance(goal, np.ndarray) and goal.shape == self.goal_shape
        self.done = False
        self.current_edge_interactions = 0
        self.last_node_reached = None
        self.sub_goals = []
        self.final_goal = goal
        if self.reachability_graph.nodes:
            self.init_path(state, self.final_goal)
        if self.sub_goals:
            self.current_goal = self.final_goal
        super().start_episode(state)
        self.control_policy.start_episode(state, self.next_goal())

    def get_node_state(self, node_id):
        return self.reachability_graph.nodes()[node_id]["state"]

    def next_goal(self):
        if self.sub_goals:
            return self.get_node_state(self.sub_goals[0])
        return self.final_goal

    def process_interaction(self, action, reward, new_state, done, learn=True):
        self.current_edge_interactions += 1
        control_agent_episode_done = False

        """
        image = self.environment.render()
        self.environment.place_point(image, self.final_goal, [255, 0, 0])
        self.environment.place_point(image, new_state, [0, 255, 0], 7)
        for sg in self.sub_goals:
            pos = self.get_node_state(sg)
            self.environment.place_point(image, self.get_node_state(sg), [0, 0, 255])
        save_image(image, self.output_directory, "img_" + str(self.episode_time_step_id))
        """
        learn = learn and not self.reachability_graph_set
        if not learn:  # We only learn at pretraining, we will not use reachability_graph there
            # Did we have sub-goals left, and did we reach the next one?
            if self.sub_goals:
                next_sub_goal = self.get_node_state(self.sub_goals[0])
                reached = self.reached(new_state, next_sub_goal)

                if reached:
                    if self.verbose:
                        print("We reached a sub-goal in ", self.current_edge_interactions, " steps", sep='')
                    self.last_node_reached = self.sub_goals.pop(0)
                    self.current_goal = self.next_goal()
                    self.current_edge_interactions = 0
                    control_agent_episode_done = True

                    # Reset a new episode for the control policy
                    self.control_policy.start_episode(new_state, self.current_goal)
                else:
                    if self.current_edge_interactions > self.max_interactions_per_edge:
                        if self.verbose:
                            print("We fail to reach the next sub-goal")
                        if self.last_node_reached is not None:
                            self.reachability_graph.get_edge_data(self.last_node_reached,
                                                                  self.sub_goals[0])["cost"] = float("inf")
                        self.done = True
            else:  # We are trying to reach the final goal after we reached every sub-goals.
                reached = (new_state == self.final_goal).all()
                if reached:
                    if self.verbose:
                        print("We reached the final goal!")
                    self.done = True
                else:
                    if self.current_edge_interactions > self.max_final_interactions:
                        if self.verbose:
                            print("We fail to reach the final goal")
                        self.done = True
        super().process_interaction(action, reward, new_state, done, learn=learn)
        self.control_policy.process_interaction(action, reward, new_state, done or self.done or control_agent_episode_done,
                                           learn=False)

    def reached(self, state: np.ndarray, goal: np.ndarray = None) -> bool:
        if goal is None:
            goal = self.get_node_state(self.sub_goals[0])
        dist = np.linalg.norm(goal - state, 2)
        return dist <= self.tolerance_radius

    def action(self, state, explore=True):
        return self.control_policy.action(state, explore=explore)

    def on_pre_training_done(self, start_state, reached_goals, oracle):
        # Compute the farthest reached goal for future q_value normalisation
        self.q_distance_max = None
        for goal in reached_goals:
            estimated_distance = self.get_distance_estimation(start_state, goal, normalised=False)
            if self.q_distance_max is None or estimated_distance > self.q_distance_max:
                self.q_distance_max = estimated_distance
        self.init_reachability_graph(oracle)

    def get_distance_estimation(self, states, goals, normalised=True):
        """
        Use the UVFA to get a value function approximation between two states.
        """
        with torch.no_grad():
            distance_estimation = self.control_policy.get_estimated_distances(states, goals)
        return distance_estimation if not normalised else distance_estimation / self.q_distance_max

    def get_position(self, state):
        if state.shape[0] == 2:
            return state
        else:
            return state[:2]

    def copy(self):
        control_agent = self.control_policy.copy()
        del self.control_policy
        new_agent = deepcopy(self)
        new_agent.control_policy = control_agent.copy()
        self.control_policy = control_agent.copy()
        return new_agent

    def reset(self):
        self.__init__(self.control_policy_goal_conditioned_wrapper, self.control_policy_value_based_agent_class,
                      self.state_space, self.action_space, **self.init_params)



from typing import Union

import networkx as nx
import numpy as np
from networkx import NetworkXNoPath

from hbrl.agents.agent import Agent
from gym.spaces import Box, Discrete

from hbrl.agents.discrete.distributional_dqn import DistributionalDQN
from hbrl.agents.goal_conditioned_wrappers.goal_conditioned_value_based_agent import GoalConditionedValueBasedAgent
from hbrl.agents.goal_conditioned_wrappers.her import HER
from hbrl.agents.continuous.distributional_ddpg import DistributionalDDPG


class SORB(Agent):

    """
    SORB agent as defined in sorb paper (Eisenbach, 2017)
    https://proceedings.neurips.cc/paper/2019/file/5c48ff18e0a47baaf81d8b8ea51eec92-Paper.pdf.
    """

    name = "SORB"

    def __init__(self, state_space: Union[Box, Discrete], action_space: Union[Box, Discrete], **params):
        """
        @param state_space: Environment's state space.
        @param action_space: Environment's action_space.
        @param params: Optional parameters.
        """

        super().__init__(state_space, action_space, **params)
        if isinstance(self.action_space, Box):
            self.control_policy: GoalConditionedValueBasedAgent = HER(DistributionalDDPG, state_space, action_space, **params)
        if isinstance(self.action_space, Discrete):
            self.control_policy: GoalConditionedValueBasedAgent = HER(DistributionalDQN, state_space, action_space, **params)
        assert hasattr(self.control_policy, "replay_buffer")
        self.current_goal = None

        # There is two phases in SORB training. First, the control policy is trained by trying to reach goals in the
        # environment. Then, we sample some states in the control policy's buffer, and build links between those that
        # are considered as reachable (check their paper for more details on that).
        # If the control policy learn again, the distance estimation (aka, the Q-value in SORB) change and the graph is
        # no more usable, because links should be computed again with the new distance function.

        self.graph_up_to_date = False
        #  '--> True if we didn't learn since the graph has been build. Set to false at each learning step.
        self.reachability_graph = nx.DiGraph()
        self.graph_building_buffer_size = params.get("nb_nodes", 1000)
        self.max_edges_length = params.get("max_edges_length", 5)
        self.reachability_threshold = params.get("reachability_threshold", 1)
        self.state_to_goal_filter = params.get("state_to_goal_filter", np.array([True, True]))
        self.sub_goals = []
        self.next_goal = None
        self.done = False
        self.nb_sub_task_interaction = 0
        self.max_interactions_per_sub_task = 20

    def reached_sub_goal(self, state, sub_goal):
        assert not isinstance(sub_goal, int)
        return np.linalg.norm(state - sub_goal) < self.reachability_threshold

    def on_pre_training_done(self):
        self.build_graph()

    def build_graph(self):
        states = self.control_policy.reinforcement_learning_agent.replay_buffer \
                    .sample(self.graph_building_buffer_size)[0][:, 2:].detach().cpu().numpy()

        for node_id, state in enumerate(states):
            self.reachability_graph.add_node(node_id, state=state)

        distances = self._get_pairwise_dist(states)

        for node_1_id, state_1 in enumerate(states[:-1]):
            for node_2_id, state_2 in enumerate(states[node_1_id + 1:]):
                node_2_id += node_1_id + 1
                dist = distances[node_1_id * states.shape[0] + node_2_id]
                if dist < self.max_edges_length:
                    self.reachability_graph.add_edge(node_1_id, node_2_id)

                dist = distances[node_2_id * states.shape[0] + node_1_id]
                if dist < self.max_edges_length:
                    self.reachability_graph.add_edge(node_2_id, node_1_id)
        self.graph_up_to_date = True

    def _get_pairwise_dist(self, states):
        """
        Estimates the pairwise distances.
        """
        assert states.shape[0] > 1, "cannot compute pair wise distance on a single state."
        nb_states = states.shape[0]
        a = np.repeat(states, nb_states, axis=0)
        b = np.tile(states, (nb_states, 1))
        pair_wise_data = np.concatenate((a, b), -1)

        batch_max_length = int(2e6)
        nb_calls = pair_wise_data.shape[0] // batch_max_length \
                   + (0 if pair_wise_data.shape[0] % batch_max_length == 0 else 1)
        q_values = None
        for i in range(nb_calls):
            #   '--> A too big batch can lead to memory issues that crash the program.
            data = pair_wise_data[i * batch_max_length: (i + 1) * batch_max_length]
            q = self.control_policy.reinforcement_learning_agent\
                .get_value(data)
            q_values = np.concatenate((q_values, q), 0) if q_values is not None else q
        return - q_values

    def get_distance_estimation(self, state_1, state_2):
        features = np.concatenate((state_1, state_2), -1)
        return - self.control_policy.reinforcement_learning_agent.get_value(features)

    def shortest_path(self, node_from, node_to_reach):
        return nx.shortest_path(self.reachability_graph, node_from, node_to_reach)

    def get_node_for_state(self, state, data=False, reachable_from=None):
        """
        Select the node that best represent the given state.
        """
        assert isinstance(state, np.ndarray)
        if state.shape[-1] == len(self.state_to_goal_filter):
            state = state[self.state_to_goal_filter]
        if not self.reachability_graph.nodes:
            return None  # The graph  hasn't been initialised yet.
        node_data = None
        closest_distance = None
        for node_id, args in self.reachability_graph.nodes(data=True):
            if reachable_from is not None:
                try:
                    self.shortest_path(reachable_from, node_id)
                except NetworkXNoPath:
                    continue
            distance = self.get_distance_estimation(args["state"], state)
            if closest_distance is None or distance < closest_distance:
                node_data = (node_id, args)
                closest_distance = distance
        return node_data if data else node_data[0]

    def start_episode(self, state: np.ndarray, goal: np.ndarray, test_episode=False):
        super().start_episode(state, test_episode=test_episode)
        self.current_goal = goal
        self.done = False
        self.nb_sub_task_interaction = 0
        if self.graph_up_to_date:
            current_node = self.get_node_for_state(state)
            final_node = self.get_node_for_state(self.current_goal, reachable_from=current_node)
            self.sub_goals = self.shortest_path(current_node, final_node)
            if self.sub_goals:
                self.next_goal = nx.get_node_attributes(self.reachability_graph, "state")[self.sub_goals[0]]
            else:
                self.next_goal = self.current_goal
        else:
            self.next_goal = self.current_goal
        self.control_policy.start_episode(state, self.next_goal, test_episode=self.under_test or self.graph_up_to_date)

    def action(self, state, explore=True):
        return self.control_policy.action(state, explore=not self.graph_up_to_date)

    def process_interaction(self, action, reward, new_state, done, learn=True):
        super().process_interaction(action, reward, new_state, done, learn=learn)
        learn = learn and not self.graph_up_to_date

        if self.graph_up_to_date:
            if self.reached_sub_goal(new_state, self.next_goal):
                if self.sub_goals:
                    self.sub_goals.pop(0)
                if self.sub_goals:
                    self.next_goal = nx.get_node_attributes(self.reachability_graph, "state")[self.sub_goals[0]]
                else:
                    self.next_goal = self.current_goal
                self.nb_sub_task_interaction = 0
                self.control_policy.start_episode(new_state, self.next_goal,
                                                  test_episode=self.under_test or self.graph_up_to_date)
            else:
                self.nb_sub_task_interaction += 1
                if self.nb_sub_task_interaction >= self.max_interactions_per_sub_task:
                    self.done = True
                assert (self.next_goal == self.control_policy.current_goal).all()
        else:
            self.control_policy.process_interaction(action, reward, new_state, done, learn=learn)
        assert (self.next_goal == self.control_policy.current_goal).all()

    def learn_interaction(self, state, action, reward, new_state, done):
        pass



import time
from copy import deepcopy


class Stopwatch:
    """
    A stopwatch that can be build using any clock. The function that will be used to get the current time
    """
    def __init__(self, get_time=time.process_time):
        assert callable(get_time)
        temp_time = time.process_time()
        self.duration = temp_time - temp_time  # Initialise to a duration of 0.
        self.last_start = None
        self.get_time = get_time

    def start(self):
        assert not self.is_running(), "Cannot start a running stopwatch."
        self.last_start = self.get_time()

    def stop(self):
        assert self.is_running(), "Stopwatch has not been started. There is probably an error."
        self.duration += self.get_time() - self.last_start
        self.last_start = None
        return self.duration

    def get_duration(self):
        if self.is_running():
            return self.duration + (self.get_time() - self.last_start)
        else:
            return deepcopy(self.duration)

    def is_running(self):
        return self.last_start is not None

    def reset(self, start=False):
        self.__init__()
        if start:
            self.start()

    def __str__(self):
        return str(self.get_duration())

if __name__ == "__main__":
    print("cpu time: ", time.process_time())

    sw = Stopwatch()
    sw.start()
    print("stated ..")
    for i in range(10000):  # Random operations to take some time. I can use a sleep() but I did like this idkw.
        a = i ** i
    sw.stop()
    print("stopped.")
    print("duration: " + str(sw) + " seconds")


import torch


def signlog(target):
    """
    This function can be applied on the critic target to help convergence.
    It scales down the critic target value, so it's no more necessary to normalise the reward.
    This function is similar to a Tanh, but it is not bounded between -1 and 1, so when the target q-value is high,
    signlog(target) can still be different enough from signlog(target - 1).

    usage:
    self.critic(state, action,

    """
    return torch.sign(target) * torch.log(target + 1)



# Goal conditioned deep Q-network
from ..value_based_agent import ValueBasedAgent
from ..utils.mlp import MLP
import copy
import numpy as np
import torch
from torch import optim
from torch.nn import ReLU


class DQN(ValueBasedAgent):
    """
    An agent that learn an approximated Q-Function using a neural network.
    This Q-Function is used to find the best action to execute in a given observation.
    """

    name = "DQN"

    def __init__(self, state_space, action_space, **params):
        """
        @param state_space: Environment's state space.
        @param action_space: Environment's action_space.
        @param params: Optional parameters.
        """

        super().__init__(state_space, action_space, **params)

        self.gamma = params.get("gamma", 0.95)
        self.epsilon_min = params.get("epsilon_min", 0.01)
        self.epsilon_max = params.get("epsilon_max", 1.)
        self.epsilon_decay_delay = params.get("epsilon_decay_delay", 20)
        self.epsilon = None
        self.epsilon_decay_period = params.get("epsilon_decay_period", 1000)

        #  NEW, goals will be stored inside the replay buffer. We need a specific one with enough place to do so
        self.learning_rate = params.get("learning_rate", 0.001)
        self.steps_before_target_update = params.get("steps_before_target_update", 1)
        self.steps_since_last_target_update = 0
        self.tau = params.get("tau", 0.001)
        self.nb_gradient_steps = params.get("nb_gradient_steps", 1)

        self.epsilon_step = (self.epsilon_max - self.epsilon_min) / self.epsilon_decay_period
        self.total_steps = 0

        # NEW, The input observation size is multiplied by two because we need to also take the goal as input
        model_layers = params.get("model_layers", [64, ReLU(), 64, ReLU(), 64, ReLU()])
        self.model = MLP(self.state_size, *model_layers, self.nb_actions,
                         learning_rate=self.learning_rate, optimizer_class=optim.Adam, device=self.device).float()

        self.criterion = torch.nn.SmoothL1Loss()
        self.target_model = copy.deepcopy(self.model).to(self.device)
        self.epsilon = self.epsilon_max

    def get_value(self, observations, actions=None):
        with torch.no_grad():
            values = self.model(observations)
            if actions is None:
                values = values.max(-1).values
            else:
                if not isinstance(actions, torch.Tensor):
                    actions = torch.tensor(actions)
                values = values.gather(1, actions.to(torch.long).unsqueeze(1))
        return values.detach().numpy()

    def action(self, state, explore=True):
        if self.simulation_time_step_id > self.epsilon_decay_delay:
            self.epsilon = max(self.epsilon_min, self.epsilon - self.epsilon_step)

        if explore and not self.under_test and np.random.rand() < self.epsilon:  # Epsilon greedy
            action = np.random.randint(self.nb_actions)
        else:
            # greedy_action(self.model, observation) function in RL5 notebook
            with torch.no_grad():
                q_values = self.model(state)
                action = torch.argmax(q_values).item()
        return action

    def learn(self):
        assert not self.under_test
        for _ in range(self.nb_gradient_steps):
            if len(self.replay_buffer) > self.batch_size:
                states, actions, rewards, new_states, dones = self.sample_training_batch()
                q_prime = self.target_model(new_states).max(1)[0].detach()
                update = rewards + self.gamma * (1 - dones) * q_prime
                q_s_a = self.model(states).gather(1, actions.to(torch.long).unsqueeze(1))
                loss = self.criterion(q_s_a, update.unsqueeze(1))
                self.model.learn(loss)

        self.steps_since_last_target_update += 1
        if self.steps_since_last_target_update >= self.steps_before_target_update:
            self.target_model.converge_to(self.model, self.tau)
            self.steps_since_last_target_update = 0



from typing import Union

import numpy as np
import importlib
import random
from gym import spaces
from matplotlib import pyplot as plt
from .utils.indexes import *
from ..maps.maps_index import MapsIndex
from scipy.spatial import distance
from skimage.draw import line_aa


class PointEnv:
    """Abstract class for 2D navigation environments."""

    def __init__(self, map_name:str = MapsIndex.EMPTY.value, action_noise=1.0, reset_anywhere=True):
        self.reset_anywhere = reset_anywhere
        self.maze_map = np.array(importlib.import_module("hbrl.environments.maps." + map_name).maze_array)
        self.height, self.width = self.maze_map.shape
        self.action_noise = action_noise
        self.state_space = spaces.Box(low=np.array([- self.width / 2, - self.height / 2]),
                                      high=np.array([self.width / 2, self.height / 2]))
        self.action_space = spaces.Box(low=np.array([-1.0, -1.0]), high=np.array([1.0, 1.0]), dtype=np.float32)
        self.agent_state = None
        self.reset()

    def get_state(self, x, y):
        """
        Return a numpy array (state) that belongs to X and Y coordinates in the grid.
        """
        x_value = x + .5 - self.width / 2
        y_value = - (y + .5 - self.height / 2)
        return np.asarray([x_value, y_value])

    def get_coordinates(self, state):
        return round(state[0].item() -.5 + self.width / 2), round(- state[1].item() - .5 + self.height / 2)

    def reset(self):
        if self.reset_anywhere:
            # Look for reachable tile in the map
            start_coordinates = np.flip(random.choice(np.argwhere(self.maze_map != 1)))
        else:
            # Look for a valid start position in the map
            start_coordinates = np.flip(random.choice(np.argwhere(self.maze_map == 2)))
        self.agent_state = self.get_state(*start_coordinates)
        if self.reset_anywhere:
            self.agent_state += np.random.uniform(low=-0.5, high=0.5, size=2)
        assert self.state_space.contains(self.agent_state.astype(self.state_space.dtype))
        return self.agent_state.copy()

    def _sample_empty_state(self):
        empty_coordinates = np.flip(random.choice(np.argwhere(self.maze_map != 1)))
        state = self.get_state(*empty_coordinates)
        state += np.random.uniform(low=-0.5, high=0.5, size=2)
        assert self.is_available(*self.get_coordinates(state))
        return state

    def get_tile_type(self, x, y):
        return TileType(self.maze_map[y][x].item())

    def is_available(self, x, y):
        # False for 218, 138
        # if we move into a row not in the grid
        if 0 > x or x >= self.width or 0 > y or y >= self.height:
            return False
        if self.get_tile_type(x, y) == TileType.WALL:
            return False
        return True

    def step(self, action):
        assert self.action_space.contains(action.astype(self.action_space.dtype))
        if self.action_noise > 0:
            action += np.random.normal(0, self.action_noise)
        action = np.clip(action, self.action_space.low, self.action_space.high)
        num_sub_steps = 10
        dt = 1.0 / num_sub_steps
        num_axis = len(action)
        for _ in np.linspace(0, 1, num_sub_steps):
            new_state = self.agent_state.copy()
            for axis in range(num_axis):
                new_state[axis] += dt * action[axis]
            if self.is_available(*self.get_coordinates(new_state)):
                self.agent_state = new_state

        done = False
        new_tile_type = self.get_tile_type(*self.get_coordinates(self.agent_state))
        reward = 10 if new_tile_type == TileType.TERMINAL else -1
        return self.agent_state.copy(), reward, done

    def get_oracle(self, step=0.2):
        oracle = []
        state_range = self.state_space.high - self.state_space.low
        x_range = int(state_range[0].item() * (1 / step))
        y_range = int(state_range[1].item() * (1 / step))
        for x in range(x_range):
            for y in range(y_range):
                state = np.array([x * step + (step / 2) + self.state_space.low[0].item(),
                                  y * step + (step / 2) + self.state_space.low[1].item()])
                oracle.append(state)
        return oracle

    """
    RENDERING FUNCTIONS
    """

    def set_tile_color(self, image_array: np.ndarray, x, y, color, tile_size=10, border_size=0) -> np.ndarray:
        """
        Set a tile color with the given color in the given image as a numpy array of pixels
        :param image_array: The image where the tile should be set
        :param x: X coordinate of the tile to set
        :param y: Y coordinate of the tile to set
        :param color: new color of the tile : numpy array [Red, Green, Blue]
        :param tile_size: size of the tile in pixels
        :param border_size: size of the tile's border in pixels
        :return: The new image
        """
        tile_img = np.zeros(shape=(tile_size, tile_size, 3), dtype=np.uint8)

        if border_size > 0:
            tile_img[:, :, :] = Colors.TILE_BORDER.value
            tile_img[border_size:-border_size, border_size:-border_size, :] = color
        else:
            tile_img[:, :, :] = color

        y_min = y * tile_size
        y_max = (y + 1) * tile_size
        x_min = x * tile_size
        x_max = (x + 1) * tile_size
        image_array[y_min:y_max, x_min:x_max, :] = tile_img
        return image_array

    def get_color(self, x, y, ignore_terminals=False):
        tile_type = self.get_tile_type(x, y)
        if tile_type == TileType.START:
            return Colors.START.value
        elif tile_type == TileType.WALL:
            return Colors.WALL.value
        elif tile_type == TileType.EMPTY:
            return Colors.EMPTY.value
        elif tile_type == TileType.TERMINAL:
            return Colors.EMPTY.value if ignore_terminals else Colors.TERMINAL.value
        else:
            raise AttributeError("Unknown tile type")

    def get_environment_background(self, tile_size=10, ignore_agent=True, ignore_rewards=False) -> np.ndarray:
        """
        Return an image (as a numpy array of pixels) of the environment background.
        :return: environment background -> np.ndarray
        """
        # Compute the total grid size
        width_px = self.width * tile_size
        height_px = self.height * tile_size

        img = np.zeros(shape=(height_px, width_px, 3), dtype=np.uint8)

        # Render the grid
        for y in range(self.height):
            for x in range(self.width):
                cell_color = self.get_color(x, y, ignore_terminals=ignore_rewards)
                img = self.set_tile_color(img, x, y, cell_color)

        if not ignore_agent:
            self.place_point(img, self.agent_state, Colors.AGENT.value, 5)

        return img

    def render(self, ignore_rewards=False):
        """
        Render the whole-grid human view
        """
        img = self.get_environment_background(ignore_agent=False, ignore_rewards=ignore_rewards)
        return img

    def place_point(self, image: np.ndarray, state, color: Union[np.ndarray, list], width=5):
        """
        Modify the input image
        param image: Initial image that will be modified.
        param x: x coordinate in the state space of the point to place.
        param y: y coordinate in the state space of the point to place.
        param color: Color to give to the pixels that compose the point.
        param width: Width of the circle (in pixels).
        """
        if isinstance(color, list):
            color = np.array(color)

        state_space_range = (self.state_space.high - self.state_space.low)
        center = (state - self.state_space.low) / state_space_range
        center[1] = 1 - center[1]
        center_y, center_x = (image.shape[:2] * np.flip(center)).astype(int)

        # Imagine a square of size width * width, with the coordinates computed above as a center. Iterate through
        # each pixel inside this square to
        radius = width
        for i in range(center_x - radius, center_x + radius):
            for j in range(center_y - radius, center_y + radius):
                dist = distance.euclidean((i, j), (center_x, center_y))
                if dist < radius and 0 <= i < image.shape[1] and 0 <= j < image.shape[0]:
                    image[j, i] = color
        return image

    def place_edge(self, image: np.ndarray, state_1, state_2, color: Union[np.ndarray, list], width=40):
        """
        Modify the input image
        param image: Initial image that will be modified.
        param x: x coordinate in the state space of the point to place.
        param y: y coordinate in the state space of the point to place.
        param color: Color to give to the pixels that compose the point.
        param width: Width of the circle (in pixels).
        """
        if isinstance(color, list):
            color = np.array(color)
        state_space_range = (self.state_space.high - self.state_space.low)

        center = (state_1 - self.state_space.low) / state_space_range
        center[1] = 1 - center[1]
        center_y_1, center_x_1 = (image.shape[:2] * np.flip(center)).astype(int)

        center = (state_2 - self.state_space.low) / state_space_range
        center[1] = 1 - center[1]
        center_y_2, center_x_2 = (image.shape[:2] * np.flip(center)).astype(int)

        rr, cc, val = line_aa(center_y_1, center_x_1, center_y_2, center_x_2)
        old = image[rr, cc]
        extended_val = np.tile(val, (3, 1)).T
        image[rr, cc] = (1 - extended_val) * old + extended_val * color

    def plot(self):
        """
        plot the environment in matplotlib windows
        """
        walls = self.maze_map.T
        (height, width) = walls.shape
        for (i, j) in zip(*np.where(walls)):
            x = np.array([j, j + 1]) / float(width)
            y0 = np.array([i, i]) / float(height)
            y1 = np.array([i + 1, i + 1]) / float(height)
            plt.fill_between(x, y0, y1, color='grey')
        plt.xlim([0, 1])
        plt.ylim([0, 1])
        plt.xticks([])
        plt.yticks([])



# Goal conditioned agent
from random import randrange
import numpy as np
from .goal_conditioned_value_based_agent import GoalConditionedValueBasedAgent


class HER(GoalConditionedValueBasedAgent):
    """
    A global agent class for goal conditioned agents. The # NEW tag indicate differences between Agent class and this
    one.
    """

    def __init__(self, reinforcement_learning_agent_class, state_space, action_space, **params):
        super().__init__(reinforcement_learning_agent_class, state_space, action_space, **params)
        self.last_trajectory = []
        # ... and store relabelling parameters
        self.nb_resample_per_states = 4
        self.name = self.reinforcement_learning_agent.name + " + HER"

    def start_episode(self, state: np.ndarray, goal: np.ndarray, test_episode=False):
        self.last_trajectory = []
        return super().start_episode(state, goal, test_episode)

    def process_interaction(self, action, reward, new_state, done, learn=True):
        if learn and not self.under_test:
            self.last_trajectory.append((self.last_state, action))
        super().process_interaction(action, reward, new_state, done, learn=learn)

    def stop_episode(self):
        # Relabel last trajectory
        if self.under_test or len(self.last_trajectory) <= self.nb_resample_per_states:
            return

        # For each observation seen :
        for state_index, (state, action) in enumerate(self.last_trajectory[:-self.nb_resample_per_states]):
            new_state_index = state_index + 1
            new_state, _ = self.last_trajectory[new_state_index]

            # sample four goals in future states
            for relabelling_id in range(self.nb_resample_per_states):
                goal_index = randrange(new_state_index, len(self.last_trajectory))
                target_state, _ = self.last_trajectory[goal_index]
                goal = target_state[self.state_to_goal_filter]

                features = self.get_features(state, goal)
                # Compute a reward that goes from -1, for the first state of the fake trajectory, to 0, if the
                # new_state if the fake goal.
                reward = (new_state_index / goal_index) - 1
                new_features = self.get_features(new_state, goal)
                done = goal_index == new_state_index
                self.reinforcement_learning_agent.save_interaction(features, action, reward, new_features, done)
        super().stop_episode()
